import sys
import __builtin__
from array import array as arr

from pds4_exceptions import PDS4StandardsException
from pds4_logging import logger_init

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    np = None

# Initialize the logger
logger = logger_init()

#################################


# Returns array struct_type, which is a tuple of (byte-order, array.array typecode, numpy typecode)
def get_array_struct_type(data_type):

    struct_type = {'IEEE754MSBSingle': ('big', 'f', 'float32'),
                   'IEEE754MSBDouble': ('big', 'd', 'float64'),
                   'SignedMSB2': ('big', 'h', 'int16'),
                   'SignedMSB4': ('big', 'i', 'int32'),
                   'SignedMSB8': ('big', 'l', 'int64'),
                   'UnsignedMSB2': ('big', 'H', 'uint16'),
                   'UnsignedMSB4': ('big', 'I', 'uint32'),
                   'UnsignedMSB8': ('big', 'L', 'uint64'),
                   'ComplexMSB8': ('big', 'f', 'float32'),
                   'ComplexMSB16': ('big', 'd', 'float64'),

                   'IEEE754LSBSingle': ('little', 'f', 'float32'),
                   'IEEE754LSBDouble': ('little', 'd', 'float64'),
                   'SignedLSB2': ('little', 'h', 'int16'),
                   'SignedLSB4': ('little', 'i', 'int32'),
                   'SignedLSB8': ('little', 'l', 'int64'),
                   'UnsignedLSB2': ('little', 'H', 'uint16'),
                   'UnsignedLSB4': ('little', 'I', 'uint32'),
                   'UnsignedLSB8': ('little', 'L', 'uint64'),
                   'ComplexLSB8': ('little', 'f', 'float32'),
                   'ComplexLSB16': ('little', 'd', 'float64'),

                   'SignedByte': ('<', 'b' 'int8'),
                   'UnsignedByte': ('<', 'B', 'uint8')
                   # 'SignedBitString': ('', ''),   # Currently unhandled
                   # 'UnsignedBitString': ('', '')

                   }.get(data_type, None)

    return struct_type


# Returns a tuple containing the smallest (array.array integer typecode, numpy integer typecode) that can
# store every value in data. Returns None if no typecode can fit every integer. Data must contain only integers.
def get_min_integer_struct_type(data):

    data_min = min(data)
    data_max = max(data)
    abs_max = max(abs(data_min), abs(data_max))

    if abs_max <= 127:
        typecode = ('b', 'int8')

    elif abs_max <= 255:

        if data_min >= 0:
            typecode = ('B', 'uint8')
        else:
            typecode = ('h', 'int16')

    elif abs_max <= 32767:
        typecode = ('h', 'int16')

    elif abs_max <= 65535:

        if data_min >= 0:
            typecode = ('H', 'uint16')
        else:
            typecode = ('i', 'int32')

    elif abs_max <= 2147483647:
        typecode = ('i', 'int32')

    elif abs_max <= 4294967295:

        if data_min >= 0:
            typecode = ('I', 'uint32')
        else:
            typecode = ('l', 'int64')

    elif abs_max <= 9223372036854775807:
        typecode = ('l', 'int64')

    elif (abs_max <= 18446744073709551615) and (data_min >= 0):
        typecode = ('L', 'uint64')

    else:
        typecode = (None, None)

    # On Windows 'l' and 'L' in array.array have a max size of 4 bytes instead of expected 8.
    if (typecode[0] is not None) and (typecode[0].lower() == 'l') and (arr('l').itemsize == 4 or arr('L').itemsize == 4):
        typecode = (None, typecode[1])

    return typecode


# Converts byte_string to array.array or numpy array having the proper typecode for data_type
def data_type_convert_array(data_type, byte_string, use_numpy=False):

    struct_type = get_array_struct_type(data_type)
    if struct_type is None:
        raise PDS4StandardsException('Found invalid data type in array data: {0}'.format(data_type))

    # Extract the byte order and type codes for python and numpy
    byte_order = struct_type[0]
    python_tc = struct_type[1]
    numpy_tc = struct_type[2]

    # Unpack byte array data into an array
    if use_numpy:
        byte_string = np.fromstring(byte_string, dtype=numpy_tc)
    else:
        byte_string = arr(python_tc, byte_string)

    # Byte swap data if necessary
    need_byteswap = False

    if (byte_order == 'little' and sys.byteorder == 'big') or (byte_order == 'big' and sys.byteorder == 'little'):
        need_byteswap = True

    if need_byteswap:

        if use_numpy:
            byte_string.byteswap(True)
        else:
            byte_string.byteswap()

    return byte_string


# Converts data (a list of byte_strings) to array.array or numpy array having the proper typecode for data_type
def data_type_convert_table_ascii(data_type, data, use_numpy=False):

    # Array struct_type, which is a tuple of (array.array type-code, numpy type-code)
    struct_type = {'ASCII_Real': ('d', 'float64'),
                   'ASCII_Integer': ('l', 'int'),
                   'ASCII_NonNegative_Integer': ('L', 'int'),
                   'ASCII_Boolean': ('b', 'bool_'),
                   'UTF8_String': ('u', 'unicode'),
                   'ASCII_Numeric_Base2': ('L', 'int'),
                   'ASCII_Numeric_Base8': ('L', 'int'),
                   'ASCII_Numeric_Base16': ('L', 'int'),
                  }.get(data_type, ('c', 'str'))

    python_tc = struct_type[0]
    numpy_tc = struct_type[1]

    # Special handling for UTF-8 string
    if data_type == 'UTF8_String':

        for i, datum in enumerate(data):
            data[i] = unicode(datum, 'utf-8')

    # Special handling for boolean due to e.g. bool('false') = True
    elif data_type == 'ASCII_Boolean':

        for i, datum in enumerate(data):

            if datum in ['true', '1']:
                data[i] = True
            else:
                data[i] = False

    # Special handling for numeric ascii types
    else:

        numeric_ascii_type = {'ASCII_Real': ('float', ),
                              'ASCII_Integer': ('int', ),
                              'ASCII_NonNegative_Integer': ('int', ),
                              'ASCII_Numeric_Base2': ('int', 2),
                              'ASCII_Numeric_Base8': ('int', 8),
                              'ASCII_Numeric_Base16': ('int', 16)
                              }.get(data_type, None)

        if numeric_ascii_type is not None:

            cast_func = getattr(__builtin__, numeric_ascii_type[0])
            args = numeric_ascii_type[1:]

            for i, datum in enumerate(data):
                data[i] = cast_func(datum, *args)

    # Convert to numpy array
    if use_numpy:

        # All integer types are unbounded in PDS4. We cast them to the minimum storage size for memory
        # efficiency, or leave them as python int or long if they are larger than 64-bit
        if (numpy_tc == 'int') and (data_type):
            numpy_tc = get_min_integer_struct_type(data)[1]

        data = np.asarray(data, dtype=numpy_tc)

    # Convert to array.array
    else:

        # For strings/unicode we leave the data as a list
        if (python_tc != 'c') and (python_tc != 'u'):

            # All integer types are unbounded in PDS4. We cast them to the minimum storage size for memory
            # efficiency, or leave them as python int or long if they are larger than 64-bit
            if python_tc.lower() == 'l':
                python_tc = get_min_integer_struct_type(data)[0]

            # Convert in-bounds integers and floats, leave everything else as list/python default types
            if python_tc is not None:
                data = arr(python_tc, data)

    # Emit memory efficiency warning if necessary
    if (use_numpy and numpy_tc is None) or (not use_numpy and python_tc is None):

        numpy_suggest = ''
        if not use_numpy:
            numpy_suggest = ' Recommend that use_numpy be set to True if data file is large.'

        logger.warn('Detected integer Field with precision exceeding memory efficient case.' + numpy_suggest)

    return data


# Converts data (a list of byte_strings) to array.array or numpy array having the proper typecode for data_type
def data_type_convert_table_binary(data_type, data, use_numpy=False):

    binary_type = get_array_struct_type(data_type)

    # Convert character data types
    if binary_type is None:
        data = data_type_convert_table_ascii(data_type, data, use_numpy)

    # Convert binary data types
    else:

        # Join data list back into a byte_string
        byte_string = ''.join(data)

        data = data_type_convert_array(data_type, byte_string, use_numpy)

    return data


# If array does not have a type code matching or big enough to fit changes by scaling_factor or
# value_offset, convert array to match
def adjust_array_type(array, scaling_factor, value_offset, use_numpy=False):

    # This method is unnecessary for lists containing python's default numeric types because python
    # will automatically perform the type conversion if needed
    if isinstance(array, list):
        return array

    scaling_is_float = isinstance(scaling_factor, float)
    offset_is_float = isinstance(value_offset, float)

    if use_numpy:
        data_is_float = str(array.dtype) == 'float64'
    else:
        data_is_float = array.typecode == 'd'

    # Set to double
    if (scaling_is_float or offset_is_float) and (not data_is_float):

        if use_numpy:

            try:
                array = array.astype('float64', copy=False)
            except TypeError:
                array = array.astype('float64')

        else:
            array = arr('d', array)

    # Set to int
    else:

        # Find minimum and maximum values of new data
        if scaling_factor is None:
            scaling_factor = 1

        if value_offset is None:
            value_offset = 0

        min_data = int(min(array)) * scaling_factor + value_offset
        max_data = int(max(array)) * scaling_factor + value_offset

        # For numpy, adjust to new data type if necessary
        if use_numpy:

            new_typecode = get_min_integer_struct_type([min_data, max_data])[1]

            if new_typecode is None:
                new_typecode = np.asarray([min_data, max_data]).dtype

                logger.warn('Detected integer Field with precision exceeding memory efficient case.')
            else:
                new_typecode = np.dtype(new_typecode)

            has_new_type = str(new_typecode) != str(array.dtype)
            no_precision_loss = (new_typecode.itemsize >= array.dtype.itemsize) or (str(new_typecode) == 'object')

            if has_new_type and no_precision_loss:

                try:
                    array = array.astype(new_typecode, copy=False)
                except TypeError:
                    array = array.astype(new_typecode)

        # For array.array, adjust to new data type if necessary
        else:

            new_typecode = get_min_integer_struct_type([min_data, max_data])[0]

            if new_typecode is None:
                array = list(array)

                logger.warn('Detected integer Field with precision exceeding memory efficient case.'
                            ' Recommend that use_numpy be set to True if data file is large.')

            else:

                has_new_type = new_typecode != array.typecode
                no_precision_loss = arr(new_typecode).itemsize >= array.itemsize

                if has_new_type and no_precision_loss:
                    array = arr(new_typecode, array)

    return array


# Assumes an array with length evenly divisible by two is passed in, where each two sequential values are the
# real and imaginary parts of a single complex value. Creates a complex() type if use_numpy is off, and a
# numpy complex type otherwise
def adjust_complex_type(array, use_numpy=False):

    if use_numpy:
        new_array_like = array[::2] + 1j * array[1::2]

    else:
        new_array_like = [complex(array[i], array[i + 1]) for i in range(0, len(array), 2)]

    return new_array_like

