import os

from pds4_general_objects import StructureList, Label
from pds4_read_tables import get_table
from pds4_read_arrays import get_array
from pds4_helpers import find_xml_element
from pds4_logging import logger_init

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    np = None

# Initialize the logger
logger = logger_init()

#################################


def pds4_read(filename, quiet=False, use_numpy=True, structure_num=None, structure_name=None, structure_lid=None):
    """ Reads PDS4 compliant data into a `StructureList`

    Given a PDS4 label, reads the PDS4 data described in the label and
    associated label meta data into an `StructureList`, with each PDS4 data
    structure (e.g. Array_2D, Table_Binary, etc) as its own `Structure`. By
    default all data structures described in the label are read-in.

    NOTES:

        Currently supports Array structures, Table_Character and Table_Binary.
        Packed bit fields in Table_Binary are not yet supported, all other
        features of previously mentioned structures are fully supported.

    Parameters:

        filename: str
            The filename, including full or relative path if necessary, of
            the PDS4 label describing the data.
        quiet: bool, optional
            Suppresses all info/warnings from being output.
        use_numpy: bool, optional
            Returned data will be an ndarray and use NumPy data types. On
            by default if NumPy is installed.
        structure_num: integer, optional
            Instead of reading all data structures, only read the n^th
            structure, where n = structure_num.
        structure_name: str, optional
            Instead of reading all data structures, only read the structure
            with a name equal to structure_name.
        structure_lid: str, optional
            Instead of reading all data structures, only read the structure
            with a local identifier equal to structure_lid.

    Returns:

        `StructureList`
            Contains PDS4 structures and label data. Can be treated/accessed/used
            like a `dict` or `list`.

    Example usage:

        Below we document how to read data described by an example label
        which has two data structures, an Array_2D_Image and a Table_Binary.
        An outline of the label, including the array and a table with 3
        fields, is given.

        >>> struct_list = pds4_read('/path/to/Example_Label.xml')

        Example Label Outline:

            Array_2D_Image: unnamed
            Table_Binary: Observations
                Field: order
                Field: wavelength
                Group: unnamed
                    Field: pos_vector

        All below documentation assumes that the above outlined label,
        containing an array that does not have a name indicated in the label,
        and a table has the name 'Observations' with 3 fields as shown
        has been read-in.

        Accessing Example Structures:

            To access the data structures in `StructureList`, which is returned by
            pds4_read(), you may use any combination of `dict` or `list`.

            >>> unnamed_array = struct_list[0]
            >>>              or struct_list['ARRAY_0']

            >>> obs_table = struct_list[1]
            >>>          or struct_list['Observations']

        Label or Structure Overview:

            To see a summary of the data structures, which for Arrays shows the
            type and dimensions of the array, and for Tables shows the type
            and number of fields, you may use the info() method. Calling
            info() on the `Structure` instead of `StructureList` will provide
            a more detailed summary, including all Fields for a table.

            >>> struct_list.info()
            >>> unnamed_array.info()
            >>> obs_table.info()

        Accessing Example Label data:

            To access the read-in data, as an array-like (either list,
            array.array or ndarray), you can use the data attribute for a
            PDS4 Array data structure, or the field() method to access a field
            for a table.

            >>> unnamed_array.data
            >>> obs_table.field('wavelength')
            >>> obs_table.field('pos_vector')

        Accessing Example Label meta data:

            You can access all meta data in the label for a given PDS4 data
            structure or field via the `OrderedDict` meta_data attribute. The
            below examples use the 'description' element.

            >>> unnamed_array.meta_data['description']

            >>> obs_table.field('wavelength').meta_data['description']
            >>> obs_table.field('pos_vector').meta_data['description']

        Accessing Example Label:

            The XML for a label is also accessible via the label attribute,
            either the entire label or for each PDS4 data structure.

            Entire label:
                >>> struct_list.label

            Part of label describing Observations table:
                >>> struct_list['Observations'].label
                >>> struct_list[1].label

            The returned object is similar to an ElementTree instance. It is
            searchable via find() and findall() methods and XPATH. Consult
            ElementTree manual for more details. For example,

            >>> struct_list.label.findall('.//disp:Display_Settings')

            Will find all elements in the entire label named 'Display_Settings'
            which are in the 'disp' namespace. You can additionally use the
            to_dict() and to_string() methods.

    """

    # Set logger to only emit error messages if requested
    if quiet:
        logger.quiet()

    # Initialize the log recording
    record_start = len(logger.get_handler('log_handler').records)

    # Disable use_numpy if NumPy is not installed
    if np is None:
        use_numpy = False

    # Read-in the PDS4 label
    logger.info('Processing label: ' + filename)

    label = Label()
    label.load(filename)

    # Extract directory of label
    label_dir = os.path.dirname(filename)

    # Variable that stores byte data for each data file_xml referenced in the label
    byte_data = []

    # Loop over each File Area Observation and read-in all relevant data files
    for i, file_xml in enumerate(label.findall('File_Area_Observational', return_ET=True)):

        data_filename = find_xml_element(file_xml, './/file_name')
        data_filepath = os.path.join(label_dir, data_filename)

        byte_limits = None

        # Check for specific structure if that is being requested
        for j, structure_xml in enumerate(file_xml):

            # Skip 'File' as these are not actually data structures
            if structure_xml.tag == 'File':
                continue

            # Find the requested structure
            offset = int(find_xml_element(structure_xml, 'offset'))

            next_offset = -1
            if j < len(file_xml) - 1:
                next_offset = int(find_xml_element(file_xml[j + 1], 'offset'))

            if structure_num is not None and structure_num == j - 1:
                byte_limits = {'from': offset, 'to': next_offset}
                break

            name = find_xml_element(structure_xml, 'name')
            lid = find_xml_element(structure_xml, 'local_identifier')

            if structure_name is not None and structure_name == name:
                structure_num = j - 1
                byte_limits = {'from': offset, 'to': next_offset}
                break

            if structure_lid is not None and structure_lid == lid:
                structure_num = j - 1
                byte_limits = {'from': offset, 'to': next_offset}
                break

        # Raise error if we were not able to find requested structure, if one was requested
        if byte_limits is None and (
                (structure_num is not None) or (structure_name is not None) or (structure_lid is not None)):
            raise ValueError('Requested PDS4 data structure does not exist in label')

        logger.info('Now reading data file_xml: ' + data_filename)

        # Read in the data for this file_xml in binary form into byte_data
        try:
            file_handler = open(data_filepath, 'rb')

            # Read-in when looking for a specific structure
            if byte_limits is not None:
                file_handler.seek(byte_limits['from'])

                # Read to from beginning of this structure to the beginning of the next structure
                if byte_limits['to'] > 0:
                    read_size = byte_limits['to'] - byte_limits['from']
                    file_byte_data = file_handler.read(read_size)

                # Read to end of file (the structure being read is the last one)
                else:
                    file_byte_data = file_handler.read()

            # Read-in all structures
            else:
                file_byte_data = file_handler.read()

            file_handler.close()

            byte_data.append(file_byte_data)
        except IOError:
            raise IOError("Unable to read data file '" + data_filepath + "' found in label")

    # Convert byte_data to tuple (in order to a pass a non-mutable variable into pds4_read_structures)
    byte_data = tuple(byte_data)

    # Extract all the PDS4 data structures specified in this label from the byte_data
    structures = pds4_read_structures(label, byte_data, use_numpy=use_numpy, structure_num=structure_num)

    # Save the log recording
    log = logger.get_handler('log_handler').to_string(record_start)

    # Create the structure list (which stores all the read-in structures, the label and the log created while
    # reading structures)
    structure_list = StructureList(structures, label, log)

    if len(structure_list) == 0:
        raise ValueError('No viable PDS4 data structures found')

    return structure_list


# Returns a list containing all read-in PDS4 data structures
def pds4_read_structures(label, file_byte_data, use_numpy=False, structure_num=None):
    structures = []
    table_types = ['Table_Character', 'Table_Binary']  # This should be removed once all table types supported

    # Storage for number of structures types added
    num_structures = {'array': 0, 'table': 0}

    # Loop over each File Area Observation
    for i, file_label in enumerate(label.findall('File_Area_Observational')):

        data_filename = find_xml_element(file_label, './/file_name')

        # Loop over each data structures inside the File Area
        for j, structure_label in enumerate(file_label):

            structures_xml = structure_label.getroot()

            # Determine structure is a table or an array
            if 'Array' in structures_xml.tag:
                structure_type = 'array'
            elif structures_xml.tag in table_types:
                structure_type = 'table'
            else:
                # Skip over structure if its not actually a supported data structure
                continue

            # Increment structure type counter
            num_structures[structure_type] += 1

            # Skip this structure if we're looking for a particular structure number and this one does not 
            # match (j-1 is to account for File element inside of file_xml)
            if (structure_num is not None) and (structure_num != j - 1):
                continue

            # Create the structure
            if structure_type == 'array':
                structure = get_array(label, structure_label, file_byte_data[i], data_filename, use_numpy, structure_num)

            elif structure_type == 'table':
                structure = get_table(structure_label, file_byte_data[i], data_filename, use_numpy, structure_num)

            # Set an ID for the structure if it has neither a local identifier or name in the label
            if structure.id is None:
                structure.id = '{0}_{1}'.format(structure_type.upper(), num_structures[structure_type] - 1)

            structures.append(structure)

    return structures


if __name__ == '__main__':

    # Safe import of argparse
    try:
        import argparse
    except ImportError:
        import pds4_tools.argparse_26 as argparse

    # Create program arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", help="Filename, including full path, of the label")
    parser.add_argument("--use_numpy", help="Returned data will be an ndarray and use numpy data types. On"
                                            " by default if numpy is installed.",
                        action="store_true", default=True)
    parser.add_argument("--quiet", help="Suppresses all info/warnings", action="store_true", default=False)
    parser.add_argument("--structure_num", type=int,
                        help="Only reads the data structure specified by zero-based order (integer)")
    parser.add_argument("--structure_name", help="Only reads the data structure specified by name")
    parser.add_argument("--structure_lid", help="Only reads the data structure specified by local identifier")

    args = parser.parse_args()

    pds4_read(args.filename, use_numpy=args.use_numpy, quiet=args.quiet, structure_num=args.structure_num,
              structure_name=args.structure_name, structure_lid=args.structure_lid)
