import xml.etree.ElementTree as ET
from xml.parsers.expat import ExpatError

from pds4_constants import PDS4_NAMESPACES
from pds4_logging import logger_init

# Safe import of ParseError (Python 2.7 only)
try:
    from xml.etree.ElementTree import ParseError

except ImportError:
    ParseError = None

# Initialize the logger
logger = logger_init()

#################################


def read_label(filename, strip_extra_whitespace=True, enforce_default_prefixes=False):

    """ Reads a PDS4 XML Label into an ElementTree object

        Parameters:

            strip_extra_whitespace: bool
                Strips whitespaces at the beginning and end of element and attribute values
            enforce_default_prefixes: bool
                Ensures the default namespace is the pds4 schema, ensures that 'disp' is the display dictionary
                prefix and 'sp' is the spectral dictionary prefix

        Returns:

            label_xml_root: ElementInterface
                Root element for read-in PDS4 label
    """

    # Read-in XML tree
    try:

        xml_tree = ET.iterparse(filename, events=['start-ns', 'start', 'end'])
    except IOError:
        raise IOError('Unable to locate or read label file: ' + filename)

    # Adjust XML tree
    try:

        node_num = 0
        namespace_map = {}

        for event, elem in xml_tree:

            # Add namespace to the namespace map
            if event == 'start-ns':

                if enforce_default_prefixes:

                    # Ensure the display dictionary has prefix 'disp'
                    if elem[1] == PDS4_NAMESPACES['disp']:
                        elem = ('disp', elem[1])

                    # Ensure the spectral dictionary has prefix 'sp'
                    if elem[1] == PDS4_NAMESPACES['sp']:
                        elem = ('sp', elem[1])

                # Add namespace to map, skipping if its a default namespace and does not match pds4 namespace
                if (elem[0] != '') and (elem[1] != PDS4_NAMESPACES['pds4']):
                    namespace_map[elem[1]] = elem[0]

                continue

            # Set the node number (all other values use the end events to ensure full element contents are populated)
            if event == 'start':
                node_num += 1
                elem.node_number = node_num
                continue

            # Strip default PDS4 namespace if requested
            if (enforce_default_prefixes) and (PDS4_NAMESPACES['pds4'] in elem.tag):
                elem.tag = elem.tag.split('{', 1)[0] + elem.tag.split('}', 1)[1]

            # Strip whitespace in elements and attributes if requested
            if strip_extra_whitespace:

                subiter = ET.ElementTree(elem).getiterator()
                attribs = elem.attrib.iteritems()

                # Strip whitespaces in elements that do not have children
                if len(elem) == 0:
                    for elem_content in subiter:

                        if elem_content.text:
                            elem_content.text = elem_content.text.strip()

                # Strip whitespaces in attributes
                for name, value in attribs:
                    elem.attrib[name] = value.strip()

        label_xml_root = xml_tree.root
        label_xml_root.namespace_map = namespace_map

    except (ExpatError, ParseError):
        raise ExpatError('The requested label file does not appear contain valid XML: ' + filename)

    return label_xml_root