from math import log
import array

from pds4_table_objects import TableStructure, TableManifest, Meta_TableStructure, Meta_FieldUniformlySampled
from pds4_data_types import (data_type_convert_table_ascii, data_type_convert_table_binary,
                             adjust_array_type, adjust_complex_type)
from pds4_read_arrays import adjust_data_values
from pds4_logging import logger_init

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    np = None

# Initialize the logger
logger = logger_init()

#################################


# Creates an array-like for storing all data of a single field (nested for fields inside of groups)
def _create_field_data_array(array_dimensions, array_example):

    if len(array_dimensions) == 1:

        if isinstance(array_example, list):
            return [None] * array_dimensions[0]

        elif isinstance(array_example, array.array):
            return array.array(array_example.typecode, [0] * array_dimensions[0])

        else:
            return np.asarray([0] * array_dimensions[0], dtype=array_example.dtype)

    else:
        next_dims = array_dimensions[1:]
        return [_create_field_data_array(next_dims, array_example) for i in xrange(0, array_dimensions[0])]


# Creates the storage array and populates it for field Uniformly_Sampled, modifies table_structure to contain
# this field
def _make_uniformly_sampled(table_structure, use_numpy=False):

    uniformly_sampled_xml = table_structure.label.find('Uniformly_Sampled')
    uniformly_sampled_data = []

    if uniformly_sampled_xml is not None:

        uni_sampled = Meta_FieldUniformlySampled()
        uni_sampled.load(uniformly_sampled_xml)

        if ('scale' not in uni_sampled) or (uni_sampled['scale'] == 'Linear'):
            current_value = uni_sampled['first_value']

            while current_value <= uni_sampled['last_value']:
                uniformly_sampled_data.append(current_value)
                current_value += uni_sampled['interval']

        elif uni_sampled['scale'] == 'Logarithmic':

            x = log(uni_sampled['first_value']) / log(10)
            current_value = uni_sampled['first_value']
            counter = 0

            while current_value < uni_sampled['last_value']:
                current_value = 10 ** (uni_sampled['interval'] * counter + x)
                counter += 1

                uniformly_sampled_data.append(current_value)

        elif uni_sampled['scale'] == 'Exponential':

            x = 10. ** uni_sampled['first_value']
            current_value = uni_sampled['first_value']
            counter = 0

            while current_value < uni_sampled['last_value']:
                current_value = log(uni_sampled['interval'] * counter + x) / log(10)
                counter += 1

                uniformly_sampled_data.append(current_value)

        if use_numpy:
            uniformly_sampled_data = np.asanyarray(uniformly_sampled_data)

        table_structure.add_field(uniformly_sampled_data, uni_sampled, use_numpy)


# Extracts and properly formats the data for a single table, modifies table_structure to contain
# all extracted fields
def _extract_table_data(file_byte_data, table_structure, table_manifest, use_numpy=False, ignore_offset=False):

    # Add Uniformly_Sampled to the table if it exists
    _make_uniformly_sampled(table_structure, use_numpy)

    # Obtain the TableManifest containing only the fields and groups as given in the schema
    # (prior to cloning of repetitions)
    initial_table = table_manifest.get_initial_manifest()

    # This for loop creates the majority of 'data' variable, which will store all columns and their values for
    # the table that was just read in. It inserts each field (and each repetition for group fields) one at a time
    for i in range(0, initial_table.num_items):

        # Skip groups; we need the i to be the index in table_manifest below but do not need groups
        if initial_table[i].is_group():
            continue

        # Retrieve final array-like dimensions for this field
        array_dimensions = [table_structure.meta_data['records']]
        initial_field = initial_table[i]
        parent_idx = i

        for j in range(0, initial_field.group_level):

            parent_idx = initial_table.get_parent_by_idx(parent_idx, return_idx=True)
            array_dimensions.insert(1, initial_table[parent_idx]['repetitions'])

        # Create array-like that will contain the data for this Field
        extracted_data = None

        # Loop over each repetition of the current Field (potentially multiple repetitions if inside a group)
        # and save each repetition's data
        matching_fields = table_manifest.get_field_by_full_name(initial_field.full_name())

        for field in matching_fields:

            # Storage for the field's data
            extracted_data_single = []
            extracted_bytes = []

            # Obtain offset for the data
            offset = 0
            if not ignore_offset:
                offset = table_structure.meta_data['offset']

            # Pre-extract necessary variables to speed up computation time
            field_location = field['location'] - 1
            record_length = table_structure.meta_data.record['record_length']
            field_length = field['length']

            # Extract the byte data for this specific column from byte_data
            for j in range(0, table_structure.meta_data['records']):
                start_idx = offset + field_location + (j * record_length)
                end_idx = start_idx + field_length

                extracted_bytes.append(file_byte_data[start_idx:end_idx])

            # Cast the byte data into appropriate data types
            try:
                if table_structure.type == 'Table_Character':
                    extracted_data_single = data_type_convert_table_ascii(field['data_type'], extracted_bytes, use_numpy)

                elif table_structure.type == 'Table_Binary':
                    extracted_data_single = data_type_convert_table_binary(field['data_type'], extracted_bytes, use_numpy)

            except ValueError as e:
                raise ValueError("Unable to convert field '{0}' to data_type '{1}': {2}"
                                 .format(field['name'], field['data_type'], e))

            # Make scaling_factor and value_offset adjustments
            scaling_factor = field.get('scaling_factor')
            value_offset = field.get('value_offset')

            if (scaling_factor is not None and scaling_factor != 1) or (value_offset is not None and value_offset != 0):

                # If 'scaling_factor' or 'value_offset' exist then we need to adjust array type if types do not match
                extracted_data_single = adjust_array_type(extracted_data_single, scaling_factor, value_offset, use_numpy)

                # Adjust data values to account for 'scaling_factor' and 'value_offset'
                adjust_data_values(extracted_data_single, scaling_factor, value_offset, use_numpy)

            # Adjust data to use complex data types if it had them
            if 'complex' in field['data_type']:
                extracted_data_single = adjust_complex_type(extracted_data_single, use_numpy)

            # Obtain a list which describes which repetition this field is in for all its parent groups
            repetitions = field.get_repetitions()
            num_repetitions = len(repetitions)

            # If Field is not inside a Group, then there is only a single matching_field and extracted data
            # for it should be the final extracted data
            if num_repetitions == 0:
                extracted_data = extracted_data_single

            # If Field is inside (potentially nested) Groups, use reduce to access proper location instead of
            # explicit recursion
            else:

                # Create array-like that will contain the data for this nested Field
                if extracted_data is None:
                    extracted_data = _create_field_data_array(array_dimensions, extracted_data_single)

                # Loop over each record,  then for each record use reduce to access the proper repetition
                # for each parent group this field is located in
                for record_num in range(0, table_structure.meta_data['records']):

                    # The final value we want to set is immutable. Therefore we must access the array-like
                    # prior to this final value in order to set it
                    pre_final_list = reduce(list.__getitem__, repetitions[0:num_repetitions-1], extracted_data[record_num])
                    pre_final_list[repetitions[num_repetitions-1]] = extracted_data_single[record_num]

        # Add read-in field to the TableStructure
        table_structure.add_field(extracted_data, initial_field, use_numpy)


# Create the table structure, containing label, data and meta data
def get_table(table_label, file_byte_data, data_filename, use_numpy=False, structure_num=None):

    table_xml = table_label.getroot()
    table_types = ['Table_Character', 'Table_Binary'] # This should be changed to work as in pds4_read_array

    # Skip over data structure if its not actually a Table
    if table_xml.tag not in table_types:
        raise TypeError('Attempted to get_table() on a non-table: ' + table_xml.tag)

    # Ignore offset when extracting data in the case of structure_num being specified (because the data has been
    # read starting from that offset already)
    ignore_offset = False
    if structure_num is not None:
        ignore_offset = True

    # Create the meta data structure for this table
    meta_table_structure = Meta_TableStructure()
    meta_table_structure.load(table_xml)

    # Create the data structure for this table
    table_structure = TableStructure(data_filename, table_label, meta_table_structure)

    logger.info('Now processing a {0} structure: {1}'.format(table_xml.tag, table_structure.id))

    # Stores data about the table's groups and fields
    table_manifest = TableManifest()
    table_manifest.load(table_xml)

    # Read the table from file into the table_structure variable
    _extract_table_data(file_byte_data, table_structure, table_manifest, use_numpy, ignore_offset)

    return table_structure
