import os
import functools
import traceback
import platform
import uuid
from Tkinter import Tk, Toplevel, Scrollbar, Text, Image, Menu, Frame, Label, Button, Canvas
from tkFileDialog import askopenfilename

import pds4_read
from pds4_logging import logger_init

# Safe import of numpy
try:
    import numpy as np
except ImportError:
    np = None

# Initialize the logger
logger = logger_init()

#################################


class PDS4Viewer(object):
    """ Main class and window manager for PDS4 Viewer

    PDS4 Viewer hides the root window, and uses additional Toplevel Windows to display its contents. Once all
    PDS4 Viewer open Windows are the closed the root window is then also closed automatically. Additionally
    the root and all other Windows can be closed anytime via .quit(). PDS4 Viewer will auto-remove any Window
    that is physically closed by the user.
    """

    def __init__(self, root):

        self._root = root
        self._root.withdraw()

        self._open_windows = []
        self.last_open_dir = None

        self._set_logo()

        # self._root.report_callback_exception = self.report_callback_exception

    # Adds window to PDS4 Viewer
    def add_window(self, window):

        self._open_windows.append(window)

        return uuid.uuid4()

    # Removes window from PDS4 Viewer
    def remove_window(self, window):

        self._open_windows = [open_window for open_window in self._open_windows if open_window.id != window.id]

        if len(self._open_windows) == 0:
            self.quit()

    # Returns true if the passed in window is open, false otherwise
    def is_window_open(self, window):

        if window in self._open_windows:
            return True

        return False

    # Quits PDS4 Viewer, closing all open windows
    def quit(self):

        # Under some circumstances destroying the root window will not close other windows, therefore
        # we do it ourselves
        for window in self._open_windows:
            window.close()

        # Under some circumstances both quit() and destroy() are necessary to kill the TK mainloop
        if self._root is not None:
            self._root.quit()
            self._root.destroy()
            self._root = None

    # def _unhandled_exception(self, *args):
    #
    #     # Should print to stderr too...
    #
    #     import traceback, tkMessageBox
    #
    #     err = traceback.format_exception(*args)
    #     tkMessageBox.showerror('Exception', err)

    # Sets logo for any default TK windows (e.g. tkFileDialog)
    def _set_logo(self):
        logo_path = os.path.join(os.path.dirname(__file__), 'logo.gif')

        if os.access(logo_path, os.R_OK):
            logo = Image('photo', file=logo_path)
            self._root.tk.call('wm', 'iconphoto', self._root._w, logo)


class Window(object):
    """ Base class of any window for PDS4 Viewer """

    def __init__(self, viewer):

        self._widget = Toplevel()
        self._widget.title('PDS4 Viewer')

        self._viewer = viewer
        self._set_logo()

        self._dependent_windows = []

        self._id = self._viewer.add_window(self)
        self._widget.protocol('WM_DELETE_WINDOW', self.close)

        self._win_dimensions = {'height': 0, 'width': 0}
        self._update_window_dimensions()

    def close(self):

        for window in self._dependent_windows:
            window.close()

        if self._viewer.is_window_open(self):
            self._widget.destroy()
            self._viewer.remove_window(self)

    @property
    def id(self):
        return self._id

    def _center_window(self):
        self._update_window_dimensions()
        x = (self._widget.winfo_screenwidth() // 2) - (self._win_dimensions['width'] // 2)
        y = (self._widget.winfo_screenheight() // 2) - (self._win_dimensions['height'] // 2)
        self._widget.geometry(
            '{0}x{1}+{2}+{3}'.format(self._win_dimensions['width'], self._win_dimensions['height'], x, y))

    def _fit_to_content(self):
        self._widget.winfo_toplevel().wm_geometry('')
        self._widget.update_idletasks()
        self._widget.geometry('{0}x{1}'.format(self._widget.winfo_width(), self._widget.winfo_height()))
        self._update_window_dimensions()

    def _set_title(self, title):
        self._widget.title(title)

    # Sets logo for the window
    def _set_logo(self):
        logo_path = os.path.join(os.path.dirname(__file__), 'logo.gif')

        if os.access(logo_path, os.R_OK):
            logo = Image('photo', file=logo_path)
            self._widget.tk.call('wm', 'iconphoto', self._widget._w, logo)

    def _update_window_dimensions(self):
        self._widget.update_idletasks()
        self._win_dimensions = {'height': self._widget.winfo_height(), 'width': self._widget.winfo_width()}

    # Returns a TK background name, automatically corrected for Macs
    def _get_bg(self, type='transparent'):

        # Return transparent (for target system) background color
        if type == 'transparent':

            if (platform.system() == 'Windows') or (platform.system() == 'Darwin'):
                bg_name = 'SystemButtonFace'

            else:
                bg_name = self._widget.cget('background')

        # Return gray background color (this is the color for most TK widgets on Windows and Linux)
        elif type == 'gray':

            if platform.system() == 'Windows':
                bg_name = 'SystemButtonFace'

            elif platform.system() == 'Darwin':
                bg_name = '#F0F0F0'

            else:
                bg_name = self._widget.cget('background')

        else:
            raise TypeError('Unknown background name.')

        return bg_name

    # Returns a TK font name, automatically corrected for Macs
    @classmethod
    def _get_font(cls, size=9, weight='', name='TkDefaultFont'):

        # Mac fonts seem to be 3 sizes too small compared to Windows and Linux
        if platform.system() == 'Darwin':
            size += 3

        return '{0} {1} {2}'.format(name, size, weight)


class StructureListWindow(Window):
    """ Window that summarizes the structures, showing some of their properties and giving buttons to open them """

    def __init__(self, viewer):

        # Set initial necessary variables and do other required initialization procedures
        super(StructureListWindow, self).__init__(viewer)

        # Create menu
        self._menu = Menu(self._widget)
        self._widget.config(menu=self._menu)

        file_menu = Menu(self._widget, tearoff=0)
        file_menu.add_command(label='Open...', command=self._open_file_box)
        file_menu.add_separator()
        file_menu.add_command(label='Exit', command=self._viewer.quit)
        self._menu.add_cascade(label='File', menu=file_menu)

        # Set window to not be resizable
        self._widget.resizable(0,0)

        # Initialize structure list window variables
        self._label_open = False

    # Opens the label, reads in any structures it contains, calls _draw_summary()
    def open_label(self, filename=None, from_existing_structures=None, quiet=False, structure_num=None,
                   structure_name=None, structure_lid=None):

        if (filename is None) and (from_existing_structures is None):
            return ValueError('Cannot open a label without either a filename or existing StructureList.')

        # Determine whether numpy was imported
        use_numpy = True
        if np is None:
            use_numpy = False

        log_start = len(logger.get_handler('log_handler').records)

        # Open the label and data, or show an error if one occurs
        try:
            if from_existing_structures is None:
                structure_list = pds4_read.pds4_read(filename, quiet=quiet, use_numpy=use_numpy,
                                                  structure_num=structure_num, structure_name=structure_name,
                                                  structure_lid=structure_lid)

                self._viewer.last_open_dir = os.path.dirname(filename)
                self._set_title("{0} - Data Structure Summary for '{1}'".format(self._widget.title(), filename))

            else:
                structure_list = from_existing_structures
                self._set_title("{0} - Data Structure Summary".format(self._widget.title()))

            self._draw_summary(structure_list)
            self._label_open = True

        except Exception:

            log = logger.get_handler('log_handler').to_string(log_start)
            ScrollTextWindow(self._viewer, 'An Error Occurred!', log + traceback.format_exc())
            # self.close()
            raise

    # Draws a summary of the opened label onto the screen
    def _draw_summary(self, structure_list):

        # Add Info menu
        info_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Info', menu=info_menu)

        info_menu.add_command(label='Show read-in log',
            command=lambda: ScrollTextWindow(self._viewer, 'Label/Data Read-in Log', structure_list.read_in_log))

        # Create main canvas (which will contain the header)
        canvas = Canvas(self._widget)
        canvas.pack(fill='both', expand=1)

        # Create canvas frame and show all structure meta data inside it
        canvas_frame = Frame(canvas, takefocus=False, height=31 * len(structure_list))
        canvas.create_window((0, 0), window=canvas_frame, anchor='nw')
        canvas_frame.grid(row=2, column=0, columnspan=4, sticky='w')

        num_buttons_max = 2

        for i, structure in enumerate(structure_list):

            index = Label(canvas_frame, text=i, font=self._get_font())
            canvas_frame.grid_columnconfigure(0, minsize=100)
            index.grid(row=i, column=0, pady=(2, 7))

            name_text = structure.id.strip()
            if len(name_text) >= 30:
                name_text = name_text[:29] + '...'

            name = Label(canvas_frame, text=name_text, font=self._get_font())
            canvas_frame.grid_columnconfigure(1, minsize=200)
            name.grid(row=i, column=1, pady=(2, 7))

            type = Label(canvas_frame, text=structure.type, font=self._get_font())
            canvas_frame.grid_columnconfigure(2, minsize=150)
            type.grid(row=i, column=2, pady=(2, 7))

            button_frame = Frame(canvas_frame)
            button_frame.grid(row=i, column=3, padx=(50, 40), sticky='w')
            font = self._get_font(weight='bold')

            open_label = functools.partial(self._open_label, structure_list.label)
            view_label = Button(button_frame, text='Label', font=font, width=7, command=open_label)
            view_label.pack(side='left')

            open_table = functools.partial(self._open_table, structure)
            view_table = Button(button_frame, text='Table', font=font, width=7, command=open_table)
            view_table.pack(side='left')

            if structure.is_array() and structure.is_displayable():
                open_image = functools.partial(self._open_image, structure)
                view_image = Button(button_frame, text='Image', font=font, width=7, command=open_image)
                view_image.pack(side='left')
                num_buttons_max += 1

        # Add the header
        header_frame = Frame(canvas, takefocus=False, height=31 * len(structure_list))
        header_frame.grid(row=0, column=0, columnspan=4, sticky='w')

        index = Label(header_frame, text='Index', font=self._get_font(10, 'bold'))
        header_frame.grid_columnconfigure(0, minsize=100)
        index.grid(row=0, column=0, pady=(2, 2))

        name = Label(header_frame, text='Name', font=self._get_font(10, 'bold'))
        header_frame.grid_columnconfigure(1, minsize=200)
        name.grid(row=0, column=1, pady=(2, 2))

        type = Label(header_frame, text='Type', font=self._get_font(10, 'bold'))
        header_frame.grid_columnconfigure(2, minsize=150)
        type.grid(row=0, column=2, pady=(2, 2))

        # Add line diving header and summary data
        line_frame = Frame(canvas, takefocus=False, height=10)
        line_frame.grid(row=1, column=0, columnspan=4, sticky='w')
        self._widget.update_idletasks()
        canvas.create_line(5, 27, self._widget.winfo_width() - 5, 27)

        # Add the View text header now that idletasks has been updated
        view = Label(header_frame, text='View', font=self._get_font(10, 'bold'))
        view_left_pad = (self._widget.winfo_width() - 100 - 200 - 150) / 2 - 13
        view.grid(row=0, column=3, pady=(2, 0), padx=(view_left_pad, 0))

    def _open_label(self, label):

        label_window = LabelWindow(self._viewer, label.to_string())

    def _open_table(self, structure):

        from pds4_table_viewer import open_table, open_table_from_array

        if 'Table' in structure.type:
            open_table(self._viewer, structure)

        elif 'Array' in structure.type:
            open_table_from_array(self._viewer, structure.data, structure.id, structure.meta_data['Element_Array']['data_type'])

        else:
            raise TypeError('Cannot show table view of structure having type ' + structure.type)

    def _open_image(self, structure):

        from pds4_image_viewer import open_image

        if structure.is_array() and structure.is_displayable():
            open_image(self._viewer, structure)

        else:
            raise TypeError('Cannot show image view of structure having type ' + structure.type)

    def _open_file_box(self):

        # On Linux, the default filetype selected goes first, but on Windows it goes last. On Mac adding
        # any type seems to only allow that type to be selected
        if platform.system() == 'Windows':
            filetypes = [('All Files', '*'), ('XML Files', '.xml')]
        elif platform.system() == 'Darwin':
            filetypes = []
        else:
            filetypes = [('XML Files', '.xml'), ('All Files', '*')]

        filename = askopenfilename(master=self._widget,
                                   filetypes=filetypes,
                                   initialdir=self._viewer.last_open_dir)

        if (filename is not None) and (filename.strip() != ''):

            if self._label_open:
                summary_window = StructureListWindow(self._viewer)
                summary_window.open_label(filename)
            else:
                self.open_label(filename)


class DataViewWindow(Window):
    """ Window for displaying a single data view; specific view windows (e.g. TabularViewWindow,
    ImageViewWindow, PlotViewWindow) extend this class """

    def __init__(self, viewer):
        # Set initial necessary variables and do other required initialization procedures
        super(DataViewWindow, self).__init__(viewer)

        # Create File Menu
        self._menu = Menu(self._widget)
        self._widget.config(menu=self._menu)

        file_menu = Menu(self._menu, tearoff=0)
        file_menu.add_command(label='Close', command=self.close)
        file_menu.add_command(label='Close All', command=self._viewer.quit)
        self._menu.add_cascade(label='File', menu=file_menu)

        # Initialize variables needed for any structure display window
        self.data = None
        self.meta_data = None
        self._settings = {}
        self._menu_options = {}
        self._data_open = False

        # Initialize window size
        width = self._widget.winfo_screenwidth() / 2
        height = self._widget.winfo_screenheight() / 2
        self._widget.geometry('{0}x{1}'.format(width, height))
        self._center_window()

        # Create the _canvas, which should contain all UI elements except the menu
        self._canvas = Canvas(self._widget, takefocus=False, bd=0, highlightthickness=0)

        # Create the _display_frame, which should contain all UI elements except the menu, the scrollbars and
        # the header/footer (if these are present)
        self._display_frame = Frame(self._canvas, takefocus=False)

        self._vert_scrollbar = None
        self._horz_scrollbar = None

    @property
    def settings(self):
        return self._settings

    # Adds a scroll event binding
    def _bind_scroll_event(self, scroll_method):

        if platform.system() == 'Linux':
            self._widget.bind("<4>", scroll_method)
            self._widget.bind("<5>", scroll_method)
        else:
            self._widget.bind("<MouseWheel>", scroll_method)


class ScrollTextWindow(Window):
    """ Window used to display messages inside a scrollable text area """

    def __init__(self, viewer, message_header, message):

        # Set initial necessary variables and do other required initialization procedures
        super(ScrollTextWindow, self).__init__(viewer)

        # Add header
        header_box = Frame(self._widget, bg=self._get_bg('gray'))
        header_box.pack(side='top', expand=1, fill='x')

        header = Label(header_box, text=message_header, bg=self._get_bg('gray'), font=self._get_font(15, 'bold'))
        header.pack(pady=10)

        # Add buttons Ok and Copy buttons below the text pad first so they remain during resizing
        button_frame = Frame(self._widget)
        button_frame.pack(pady=10, side='bottom')

        ok_button = Button(button_frame, bg=self._get_bg(), width=5, text='Ok',
                           font=self._get_font(weight='bold'), command=self.close)
        ok_button.pack(side='left', padx=(0, 5))

        copy_button = Button(button_frame, bg=self._get_bg(), width=7, text='Copy',
                             font=self._get_font(weight='bold'), command=self._copy_message)
        copy_button.pack(side='left')

        # Add text pad
        self._text_pad = None
        self._draw_text_pad()
        self._text_pad.insert('end', message)
        self._text_pad.config(state='disabled')
        self._text_pad.see('end')

        self._fit_to_content()
        self._center_window()

    def _draw_text_pad(self):
        frame = Frame(self._widget)
        self._text_pad = Text(frame, wrap='none', width=50, height=20)

        vert_scrollbar = Scrollbar(frame, orient='vertical', command=self._text_pad.yview)
        self._text_pad.configure(yscrollcommand=vert_scrollbar.set)

        horz_scrollbar = Scrollbar(frame, orient='horizontal', command=self._text_pad.xview)
        self._text_pad.configure(xscrollcommand=horz_scrollbar.set)

        vert_scrollbar.pack(side='right', fill='y')
        horz_scrollbar.pack(side='bottom', fill='x')
        self._text_pad.pack(side='left', expand=2, fill='both')
        frame.pack(expand=2, fill='both')

    def _copy_message(self):
        message = self._text_pad.get("1.0", 'end-1c')

        self._widget.clipboard_clear()
        self._widget.clipboard_append(message)


class LabelWindow(Window):
    """ Window used to display the XML content of a label """

    def __init__(self, viewer, string_label):

        # Set initial necessary variables and do other required initialization procedures
        super(LabelWindow, self).__init__(viewer)

        # Add header
        header_box = Frame(self._widget, bg=self._get_bg('gray'))
        header_box.pack(side='top', fill='x')

        header = Label(header_box, text='Label', bg=self._get_bg('gray'), font=self._get_font(15, 'bold'))
        header.pack(pady=10, side='top')

        # Add text pad
        self._text_pad = None
        self._draw_text_pad()
        self._text_pad.see('0.0')
        self._text_pad.insert('1.0', string_label)
        self._text_pad.config(state='disabled')

        self._fit_to_content()
        self._center_window()

    def _draw_text_pad(self):

        frame = Frame(self._widget)

        # Add search box (added first so it stays when window resized to be smaller)
        # search_box_frame = Frame(self._widget, height=20, width=275)
        # search_box_frame.pack_propagate(0)
        # search_box_frame.pack(side='bottom', anchor='nw', padx=(5, 0), pady=5)
        #
        # search_box = Entry(search_box_frame, bd=0, highlightthickness=0)
        # search_box.pack(side='left')
        #
        # search_button = Button(search_box_frame, text='Search', width=7) #command=open_label
        # search_button.pack(side='left', padx=(5, 0), fill='y', expand=1)
        #
        # regex_button = Checkbutton(search_box_frame, text="Regex") #variable=var
        # regex_button.pack(side='left', padx=(5, 0), fill='y', expand=1)

        # Add text pad
        self._text_pad = Text(frame, wrap='none', width=100, height=20)

        # Add scrollbars
        vert_scrollbar = Scrollbar(frame, orient='vertical', command=self._text_pad.yview)
        self._text_pad.configure(yscrollcommand=vert_scrollbar.set)

        horz_scrollbar = Scrollbar(frame, orient='horizontal', command=self._text_pad.xview)
        self._text_pad.configure(xscrollcommand=horz_scrollbar.set)

        vert_scrollbar.pack(side='right', fill='y')
        horz_scrollbar.pack(side='bottom', fill='x')
        self._text_pad.pack(side='left', expand=1, fill='both')
        frame.pack(side='top', expand=1, fill='both')


def pds4_viewer(filename=None, from_existing_structures=None, quiet=False, structure_num=None,
                structure_name=None, structure_lid=None):
    """ Displays PDS4 compliant data in a GUI

    Given a PDS4 label, displays PDS4 data described in the label and
    associated label meta data in a GUI. By default all data structures described
    in the label are read-in and displayed. Can be called without any
    parameters, opening a GUI that has a File->Open function to select
    desired label to be read-in and displayed.

    Parameters:

        filename: str, optional
            The filename, including full or relative path if necessary, of
            the PDS4 label describing the data to be viewed.
        from_existing_structures: StructureList, optional
            An existing StructureList, as returned by pds4_read(), to view. Takes
            precedence if given together with filename.
        quiet: bool, optional
            Suppresses all info/warnings from being output.
        structure_num: integer, optional
            Instead of reading all data structures, only read the n^th
            structure, where n = structure_num.
        structure_name: str, optional
            Instead of reading all data structures, only read the structure
            with a name equal to structure_name.
        structure_lid: str, optional
            Instead of reading all data structures, only read the structure
            with a local identifier equal to structure_lid.

    """

    root = Tk()
    viewer = PDS4Viewer(root)

    summary_window = StructureListWindow(viewer)
    summary_window.open_label(filename=filename, from_existing_structures=from_existing_structures,
                              quiet=quiet, structure_num=structure_num, structure_name=structure_name,
                              structure_lid=structure_lid)

    root.mainloop()


if __name__ == '__main__':

    # Safe import of argparse
    try:
        import argparse
    except ImportError:
        import pds4_tools.argparse_26 as argparse

    # Create program arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", help="Filename, including full path, of the label", nargs='?', default=None)
    parser.add_argument("--quiet", help="Suppresses all info/warnings", action="store_true", default=False)
    parser.add_argument("--structure_num", type=int,
                        help="Only reads the data structure specified by zero-based order (integer)")
    parser.add_argument("--structure_name", help="Only reads the data structure specified by name")
    parser.add_argument("--structure_lid", help="Only reads the data structure specified by local identifier")

    args = parser.parse_args()

    pds4_viewer(args.filename, structure_num=args.structure_num, structure_name=args.structure_name,
                structure_lid=args.structure_lid)
